﻿#region Licence
// MIT Licence
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.using System;
#endregion
/*

This software was built using tooling created and shared by others.

Big thanks to
 
Jakub Bartkowiak (Gralinfor) - Gralin.NETMF.Nordic.NRF24L01Plus
https://nrf24l01.codeplex.com/
 
Antao Almada - HydraMF.BitConverter
https://bitbucket.org/aalmada/hydramf/src/779d1c524a9c/HydraMF.Hardware/BitConverter.cs

Mike McCauley - NRF24 library for Arduino, I used the CrazyFlie emulator to debug my implementation. No doubt averting me crashing my Crazyflie many times while debugging.
http://www.airspayce.com/mikem/arduino/NRF24/index.html

BitCraze (http://www.bitcraze.se/) for deciding to open source the Crazyflie, 

*/
using System;
using System.Text;
using System.Threading;
using Microsoft.SPOT;
using Microsoft.SPOT.Hardware;

using Gralin.NETMF.Nordic;
using HydraMF ;
using SecretLabs.NETMF.Hardware.Netduino;



namespace devMobile.Crazyflie
{
   public class NetduinoRemoteControl
   {
      private AnalogInput pitchInput = new AnalogInput(Cpu.AnalogChannel.ANALOG_0);
      private const float PitchMinimum = -20.0f;
      private const float PitchMaximum = 20.0f;
      private const float PitchRange = ( PitchMaximum - PitchMinimum ) ;
      private float _PitchOffset = 0.0f;
      private float _PitchLatest = 0.0f;

      private AnalogInput rollInput = new AnalogInput(Cpu.AnalogChannel.ANALOG_1);
      private const float RollMinimum = -20.0f;
      private const float RollMaximum = 20.0f;
      private const float RollRange = ( RollMaximum - RollMinimum ) ;
      private float _RollOffset = 0.0f;
      private float _RollLatest = 0.0f;

      private AnalogInput yawInput = new AnalogInput(Cpu.AnalogChannel.ANALOG_4);
      private const float YawMinimum = -100.0f;
      private const float YawMaximum = 100.0f;
      private const float YawRange = ( YawMaximum - YawMinimum ) ;
      private float _YawOffset = 0.0f;
      private float _YawLatest = 0.0f;

      private AnalogInput thrustInput = new AnalogInput(Cpu.AnalogChannel.ANALOG_3);
      private const ushort ThrustMinimum = 0;
      private const ushort ThrustMaximum = (ushort)(ushort.MaxValue * 0.8);
      private const ushort ThrustRange = ThrustMaximum - ThrustMinimum;
      private float _ThrustOffset = 0.0f;
      private float _ThrustLatest = 0.0f;

      private const byte RadioChannel = 10;
      private readonly NRF24L01Plus _NRF24L01Plusmodule;
      private readonly byte[] _CrazyFlieAddress = { 0xe7, 0xe7, 0xe7, 0xe7, 0xe7 };
      private readonly byte[] _ControllerAddress = { 0xe7, 0xe7, 0xe7, 0xe7, 0xe7 };
      private const int OffsetCalculationNumberOfSamples = 100;
      private readonly OutputPort _FlightControlMessageSent = new OutputPort(Pins.ONBOARD_LED, false);

      
      
      public NetduinoRemoteControl()
      {
         _NRF24L01Plusmodule = new NRF24L01Plus();
      }



      public void Run()
      {
         _NRF24L01Plusmodule.OnDataReceived += OnReceive;
         _NRF24L01Plusmodule.OnTransmitFailed += OnSendFailure;
         _NRF24L01Plusmodule.OnTransmitSuccess += OnSendSuccess;

         _NRF24L01Plusmodule.Initialize(SPI.SPI_module.SPI1, Pins.GPIO_PIN_D10, Pins.GPIO_PIN_D9, Pins.GPIO_PIN_D1);
         _NRF24L01Plusmodule.Configure(_ControllerAddress, RadioChannel, NRFDataRate.DR250kbps);
         _NRF24L01Plusmodule.Enable();

         // Calibrate the thrust, yaw, pitch and roll joystick offsets
         Debug.Print("Joystick Calibration- start");
         for (int i = 0; i < OffsetCalculationNumberOfSamples; i++)
         {
            _ThrustOffset += (float)thrustInput.Read();
            _YawOffset += (float)yawInput.Read();
            _PitchOffset += (float)pitchInput.Read();
            _RollOffset += (float)rollInput.Read();
         }

         _ThrustOffset = _ThrustOffset / OffsetCalculationNumberOfSamples;
         _YawOffset = _YawOffset/ OffsetCalculationNumberOfSamples;
         _PitchOffset = _PitchOffset/ OffsetCalculationNumberOfSamples;
         _RollOffset = _RollOffset / OffsetCalculationNumberOfSamples;

         Debug.Print("T-" + _ThrustOffset.ToString("F1") + " Y " + _YawOffset.ToString("F1") + " P-" + _PitchOffset.ToString("F1") + " R-" + _RollOffset.ToString("F1"));
         Debug.Print("Joystick Calibration- Done");

         Timer flightControlTimer = new Timer(FlightControlSendMessage, null, new TimeSpan(0, 0, 0, 0, 250), new TimeSpan(0, 0, 0, 0, 100));

         Thread.Sleep(Timeout.Infinite);   
      }



      private void OnSendSuccess()
      {
         _FlightControlMessageSent.Write(false);
         Debug.Print("Send Success!");
      }



      private void OnSendFailure()
      {
         Debug.Print("Send failed!");
      }

      
      
      private void OnReceive(byte[] data)
      {
         string message = new String(Encoding.UTF8.GetChars(data));

         Debug.Print("Receive " + message); ;
      }



      float RollCalculator()
      {
         float roll = (float)rollInput.Read() - _RollOffset;

         roll = roll * RollRange;

         return roll;
      }

      
      
      float PitchCalculator()
      {
         float pitch = (float)pitchInput.Read() - _PitchOffset;

         pitch = pitch * PitchRange;

         return pitch;
      }


      float YawCalculator()
      {
         float yaw = ((float)yawInput.Read() - _YawOffset) / 0.6f;

         yaw = yaw * YawRange;

         return yaw;
      }


      ushort ThrustCalculator()
      {
         float thrust = ((float)thrustInput.Read() - _ThrustOffset) / 0.3f;

         thrust = (float)System.Math.Max(thrust, 0.0);

         return ((ushort)(thrust * ThrustRange));
      }



      private void FlightControlSendMessage(object state)
      {
         byte[] command = new byte[15];

         float pitch = PitchCalculator();
         float roll = RollCalculator();
         float yaw = YawCalculator();
         ushort thrust = ThrustCalculator();

         if ((_ThrustLatest != thrust) || (_YawLatest != yaw) || (_PitchLatest != pitch) || (_RollLatest != roll))
         {
            Debug.Print("Y " + yaw.ToString("F1") + " P-" + pitch.ToString("F1") + " R-" + roll.ToString("F1") + " T-" + thrust.ToString("F1"));

            _ThrustLatest = thrust;
            _YawLatest = yaw;
            _PitchLatest = pitch;
            _RollLatest = roll;
         }

         HydraMF.BitConverter.InsertValueIntoArray(command, 0, (byte)0x3C);
         HydraMF.BitConverter.InsertValueIntoArray(command, 1, roll );
         HydraMF.BitConverter.InsertValueIntoArray(command, 5, pitch);
         HydraMF.BitConverter.InsertValueIntoArray(command, 9, yaw);
         HydraMF.BitConverter.InsertValueIntoArray(command, 13, thrust);

        _NRF24L01Plusmodule.SendTo(_CrazyFlieAddress, command);

        _FlightControlMessageSent.Write(true);
      }
   }
}